/**
 * ASMethod.java
 * 
 * Copyright (c) 2006 David Holroyd
 */

package uk.co.badgersinfoil.metaas.dom;

import java.util.List;


/**
 * Represents an ActionScript method definition within an ActionScript class or
 * interface.
 * 
 * @see ASType#newMethod(String, Visibility, String)
 */
public interface ASMethod extends StatementContainer, ASMember {

	/**
	 * Constants defined by this class denote whether a method is actually
	 * an acessor 'get' or 'set' function.
	 * 
	 * @see ASMethod#getAccessorRole()
	 * @see ASMethod#setAccessorRole(AccessorRole)
	 */
	public static class AccessorRole {
		private String text;

		private AccessorRole(String text) { this.text = text; }
		/**
		 * A standard method; not a getter or setter.
		 */
		public static final AccessorRole NORMAL_METHOD = new AccessorRole("NORMAL_METHOD");
		/**
		 * The method is a 'get' accessor
		 */
		public static final AccessorRole SETTER = new AccessorRole("SETTER");
		/**
		 * The method is a 'set' accessor
		 */
		public static final AccessorRole GETTER = new AccessorRole("GETTER");
		
		public String toString() { return text; }
	}

	/**
	 * Returns the name of this method.
	 */
	public String getName();
	
	/**
	 * Returns a list of the formal arguments accepted by this ActionScript
	 * method.  Note that the resulting list is not modifiable.
	 * 
	 * @see #addParam(String, String)
	 */
	public List getArgs();

	/**
	 * Returns the name of the return type of this ActionScript
	 * method, or null if it is untyped.
	 */
	public String getType();

	/**
	 * Defines the name of the type of object returned by this ActionScript
	 * method.  May be set to null, denoting that the return value is
	 * untyped.
	 */
	public void setType(String string);

	/**
	 * Returns a value representing any protection-against-access defined
	 * for this ActionScript method.
	 */
	public Visibility getVisibility();

	/**
	 * Defines the level of protection-against-external-access for this
	 * ActionScript method.
	 */
	public void setVisibility(Visibility visibility);

	/**
	 * Returns true if this ActionScript method is static (i.e. the method
	 * definition uses the <code>static</code> keyword).
	 */
	public boolean isStatic();

	/**
	 * Defines whether this ActionScript method is static or not.
	 */
	public void setStatic(boolean s);

	/**
	 * Adds a formal parameter to the list of parameters supported by this
	 * ActionScript method.
	 * 
	 * @param name The name for the parameter
	 * @param type The type for the parameter, or null if the parameter is
	 *        to be untyped
	 * @return the newly created paremeter
	 */
	public ASArg addParam(String name, String type);

	/**
	 * Removes the named parameter from this ActionScript methods formal
	 * parameter list.
	 * 
	 * @return the removed parameter, if found, or null otherwise.
	 */
	public ASArg removeParam(String string);

	/**
	 * Returns on of {@link AccessorRole#NORMAL_METHOD},
	 * {@link AccessorRole#GETTER} or {@link AccessorRole#SETTER}, with
	 * NORMAL_METHOD being the default for newly synthesized methods.
	 */
	public AccessorRole getAccessorRole();

	/**
	 * Allows the role of a method to be changed.
	 */
	public void setAccessorRole(AccessorRole getter);
}