/**
 * StatementContainer.java
 * 
 * Copyright (c) 2006 David Holroyd
 */

package uk.co.badgersinfoil.metaas.dom;

import java.util.List;
import uk.co.badgersinfoil.metaas.SyntaxException;

/**
 * Defines the common services provided by structures which can contain
 * ActionScript 'statements'.
 */
public interface StatementContainer {
	/**
	 * Checks the syntax of the given code, and then adds the statement to
	 * the end of the current block.
	 * 
	 * @throws SyntaxException if the syntax of the given code fragment is
	 *         incorrect.
	 */
	public void addStmt(String statement);


	/**
	 * Adds a new expression-statement to the code, and returns a reference
	 * to it.  An 'expression-statement' is a statement that just contains
	 * an expression (followed in normal ActionScript by a semicolon, though
	 * that should be omitted from the supplied string).
	 */
	public ASExpressionStatement newExprStmt(String expr);

	/**
	 * Adds a single-line comment to list of statements being generated
	 * 
	 * @param text the text of the comment (minus the initial '//') which
	 *        must not include any newline characters.
	 */
	public void addComment(String text);

	/**
	 * Adds an if-statement to the code. e.g.
	 * <pre>block.newIf("test()").addStmt("trace('success')")</pre>
	 * results in
	 * <pre>
	 * if (test()) {
	 * 	trace('success');
	 * }</pre>
	 */
	public ASIfStatement newIf(String condition);

	/**
	 * Adds a C-style for-loop to the code.  e.g.
	 * <pre>block.newFor("var i=0", "i&lt;10", "i++").addStmt("trace(i)")</pre>
	 * results in
	 * <pre>
	 * for (var i=0; i&lt;10; i++) {
	 * 	trace(i);
	 * }</pre>
	 */
	public ASForStatement newFor(String init, String condition, String update);

	/**
	 * Adds a for-in-loop to the code.  e.g.
	 * <pre>block.newForIn("var i", "myArray").addStmt("trace(i)")</pre>
	 * results in
	 * <pre>
	 * for (var i in myArray) {
	 * 	trace(i);
	 * }</pre>
	 */
	public ASForInStatement newForIn(String init, String list);

	public ASForEachInStatement newForEachIn(String init, String list);

	/**
	 * Adds a while-loop to the code.  e.g.
	 * <pre>block.newWhile("test()").addStmt("trace('hi there')")</pre>
	 * results in
	 * <pre>
	 * while (test()) {
	 * 	trace('hi there');
	 * }</pre>
	 */
	public ASWhileStatement newWhile(String condition);

	/**
	 * Adds a do-while-loop to the code.  e.g.
	 * <pre>block.newDoWhile("test()").addStmt("trace('hi there')")</pre>
	 * results in
	 * <pre>
	 * do {
	 * 	trace('hi there');
	 * } while (test());</pre>
	 */
	public ASDoWhileStatement newDoWhile(String condition);

	/**
	 * Adds a switch-statement to the code. See {@link ASSwitchStatement}
	 * for more information.
	 */
	public ASSwitchStatement newSwitch(String condition);

	/**
	 * Adds a new with-statement to the code, and returns a reference to
	 * it.  e.g.
	 * <pre class="eg">method.newWith("value").addStmt("trace(test)");</pre>
	 * <p>results in,</p>
	 * <pre class="eg">with (value) {
	 * 	trace(test);
	 * }</pre>
	 */
	public ASWithStatement newWith(String string);

	/**
	 * Adds a new variable declaration to the code, and returns a reference
	 * to it.  e.g.
	 * <pre class="eg">method.newDeclaration("a=1");</pre>
	 * <p>results in,</p>
	 * <pre class="eg">var a=1;</pre>
	 */
	public ASDeclarationStatement newDeclaration(String assignment);

	/**
	 * Adds a new return-statement to the code (with optional return
	 * expression), and returns a reference to it.  e.g.
	 * <pre class="eg">method.newReturn(null);</pre>
	 * <p>results in a plain return statement,</p>
	 * <pre class="eg">return;</pre>
	 * <p>Whereas passing an expression,</p>
	 * <pre class="eg">method.newReturn("theVal()");</pre>
	 * <p>results will cause that expression to be returned,</p>
	 * <pre class="eg">return theVal();</pre>
	 */
	public ASReturnStatement newReturn(String expr);

	/**
	 * Returns true if if this container currently contains at
	 * least one statement, and false if it is empty, or contains only
	 * comments and whitespace.
	 */
	public boolean containsCode();

	/**
	 * Returns a list of the statements held in the containing element.  The
	 * list is immutable (entries cannnot be added, removed or replaced)
	 * but the objects obtained from the list my be modified via the
	 * methods they provide.
	 */
	public List getStatementList();
}