/**
 * ASTASMethod.java
 * 
 * Copyright (c) 2006-2007 David Holroyd
 */

package uk.co.badgersinfoil.metaas.impl;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import org.asdt.core.internal.antlr.AS3Parser;
import uk.co.badgersinfoil.metaas.SyntaxException;
import uk.co.badgersinfoil.metaas.dom.ASArg;
import uk.co.badgersinfoil.metaas.dom.ASDeclarationStatement;
import uk.co.badgersinfoil.metaas.dom.ASDoWhileStatement;
import uk.co.badgersinfoil.metaas.dom.ASExpressionStatement;
import uk.co.badgersinfoil.metaas.dom.ASForEachInStatement;
import uk.co.badgersinfoil.metaas.dom.ASForInStatement;
import uk.co.badgersinfoil.metaas.dom.ASForStatement;
import uk.co.badgersinfoil.metaas.dom.ASIfStatement;
import uk.co.badgersinfoil.metaas.dom.ASMethod;
import uk.co.badgersinfoil.metaas.dom.ASReturnStatement;
import uk.co.badgersinfoil.metaas.dom.ASSwitchStatement;
import uk.co.badgersinfoil.metaas.dom.ASWhileStatement;
import uk.co.badgersinfoil.metaas.dom.ASWithStatement;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTree;


public class ASTASMethod extends ASTASMember implements ASMethod {
	private ASTStatementList stmtList;

	public ASTASMethod(LinkedListTree ast) {
		super(ast);
		stmtList = new ASTStatementList(ASTUtils.findChildByType(ast, AS3Parser.BLOCK));
	}

	public String getName() {
		ASTIterator i = new ASTIterator(ast);
		LinkedListTree name = i.find(AS3Parser.IDENT);
		return name.getText();
	}

	public void setName(String name) {
		if (name.indexOf('.') != -1) {
			throw new SyntaxException("Method name must not contain '.'");
		}
		if (name.indexOf(':') != -1) {
			throw new SyntaxException("Method name must not contain ':'");
		}
		ASTIterator i = new ASTIterator(ast);
		i.find(AS3Parser.IDENT);
		LinkedListTree newName = AS3FragmentParser.parseIdent(name).getFirstChild();
		i.replace(newName);
	}

	public List getArgs() {
		List results = new LinkedList();
		LinkedListTree params = ASTUtils.findChildByType(ast, AS3Parser.PARAMS);
		for (ASTIterator i=new ASTIterator(params); i.hasNext(); ) {
			results.add(new ASTASArg(i.next()));
		}
		return Collections.unmodifiableList(results);
	}

	public String getType() {
		LinkedListTree typeSpec = ASTUtils.findChildByType(ast, AS3Parser.TYPE_SPEC);
		if (typeSpec == null) return null;
		return ASTUtils.typeSpecText(typeSpec);
	}

	public void setType(String typeName) {
		LinkedListTree type = ASTUtils.findChildByType(ast, AS3Parser.TYPE_SPEC);
		if (typeName == null) {
			if (type != null) {
				ast.deleteChild(ast.getIndexOfChild(type));
			}
			return;
		}
		LinkedListTree newType = AS3FragmentParser.parseTypeSpec(typeName);
		if (type == null) {
			ast.addChildWithTokens(ast.getChildCount()-1, newType);
		} else {
			type.setChildWithTokens(0, newType.getFirstChild());
		}
	}

	public void addStmt(String statement) {
		stmtList.addStmt(statement);
	}

	public ASExpressionStatement newExprStmt(String expr) {
		return stmtList.newExprStmt(expr);
	}

	private LinkedListTree findParams() {
		return ASTUtils.findChildByType(ast, AS3Parser.PARAMS);
	}

	public ASArg addParam(String name, String type) {
		LinkedListTree param = ASTUtils.newAST(AS3Parser.PARAM);
		param.addChildWithTokens(ASTUtils.newAST(AS3Parser.IDENT, name));
		if (type != null) {
			param.addChildWithTokens(AS3FragmentParser.parseTypeSpec(type));
		}
		LinkedListTree params = findParams();
		if (params.getChildCount() > 0) {
			params.appendToken(TokenBuilder.newComma());
			params.appendToken(TokenBuilder.newSpace());
		}
		params.addChildWithTokens(param);
		return new ASTASArg(param);
	}

	public ASArg removeParam(String name) {
		LinkedListTree params = ASTUtils.findChildByType(ast, AS3Parser.PARAMS);
		int count = 0;
		for (ASTIterator i=new ASTIterator(params); i.hasNext(); ) {
			LinkedListTree ast = i.next();
			ASTASArg arg = new ASTASArg(ast);
			if (arg.getName().equals(name)) {
				if (i.hasNext()) {
					ASTUtils.removeTrailingWhitespaceAndComma(ast.getStopToken());
				}
				i.remove();
				return arg;
			}
			count++;
		}
		return null;
	}

	public void addComment(String text) {
		stmtList.addComment(text);
	}

	public AccessorRole getAccessorRole() {
		LinkedListTree decl = ASTUtils.findChildByType(ast, AS3Parser.FUNCTION);
		LinkedListTree roleNode = decl.getFirstChild();
		if (roleNode == null) {
			return AccessorRole.NORMAL_METHOD;
		}
		switch (roleNode.getType()) {
		    case AS3Parser.GET:
			return AccessorRole.GETTER;
		    case AS3Parser.SET:
			return AccessorRole.SETTER;
		    default:
			throw new SyntaxException("expected GET or SET; got "+roleNode);
		}
	}

	public void setAccessorRole(AccessorRole role) {
		LinkedListTree decl = ASTUtils.findChildByType(ast, AS3Parser.FUNCTION);
		LinkedListTree roleNode = decl.getFirstChild();
		if (roleNode != null) {
			if (role == AccessorRole.NORMAL_METHOD) {
				decl.deleteChild(0);
			} else if (role == AccessorRole.GETTER) {
				roleNode.token.setType(AS3Parser.GET);
				roleNode.token.setText("get");
			} else if (role == AccessorRole.SETTER) {
				roleNode.token.setType(AS3Parser.SET);
				roleNode.token.setText("set");
			} else {
				throw new IllegalArgumentException("bad role: "+role);
			}
		} else {
			if (role == AccessorRole.GETTER) {
				LinkedListTree node = ASTUtils.newAST(AS3Parser.GET, "get");
				node.appendToken(TokenBuilder.newSpace());
				decl.addChildWithTokens(node);
			} else if (role == AccessorRole.SETTER) {
				LinkedListTree node = ASTUtils.newAST(AS3Parser.SET, "set");
				node.appendToken(TokenBuilder.newSpace());
				decl.addChildWithTokens(node);
			}
		}
	}

	public ASIfStatement newIf(String condition) {
		return stmtList.newIf(condition);
	}

	public ASForStatement newFor(String init, String condition, String update) {
		return stmtList.newFor(init, condition, update);
	}

	public ASForInStatement newForIn(String init, String list) {
		return stmtList.newForIn(init, list);
	}

	public ASForEachInStatement newForEachIn(String init, String list) {
		return stmtList.newForEachIn(init, list);
	}

	public ASWhileStatement newWhile(String condition) {
		return stmtList.newWhile(condition);
	}

	public ASDoWhileStatement newDoWhile(String condition) {
		return stmtList.newDoWhile(condition);
	}

	public ASSwitchStatement newSwitch(String condition) {
		return stmtList.newSwitch(condition);
	}

	public ASWithStatement newWith(String expr) {
		return stmtList.newWith(expr);
	}

	public ASDeclarationStatement newDeclaration(String assignment) {
		return stmtList.newDeclaration(assignment);
	}

	public ASReturnStatement newReturn(String expr) {
		return stmtList.newReturn(expr);
	}

	public boolean containsCode() {
		return stmtList.containsCode();
	}

	public List getStatementList() {
		return stmtList.getStatementList();
	}

	public String getDocComment() {
		return DocCommentUtils.getDocComment(ast);
	}


	public void setDocComment(String text) {
		DocCommentUtils.setDocComment(ast, text);
	}
}