/**
 * ASTBuilder.java
 * 
 * Copyright (c) 2006-2007 David Holroyd
 */

package uk.co.badgersinfoil.metaas.impl;

import org.asdt.core.internal.antlr.AS3Parser;
import uk.co.badgersinfoil.metaas.SyntaxException;
import uk.co.badgersinfoil.metaas.dom.ASCompilationUnit;
import uk.co.badgersinfoil.metaas.dom.Visibility;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListToken;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTree;


/**
 * Utilities to build and assemble Abstract Syntax Tree fragments for inclusion
 * into the compilation unit being generated.
 */
public class ASTBuilder {

	private ASTBuilder() {
		// hide default ctor
	}

	public static AS3ASTCompilationUnit synthesizeClass(String qualifiedName) {
		LinkedListTree unit = ASTUtils.newImaginaryAST(AS3Parser.COMPILATION_UNIT);
		LinkedListTree pkg = ASTUtils.newAST(AS3Parser.PACKAGE, "package");
		pkg.appendToken(TokenBuilder.newSpace());
		unit.addChildWithTokens(pkg);
		String packageName = packageNameFrom(qualifiedName);
		if (packageName != null) {
			pkg.addChildWithTokens(AS3FragmentParser.parseIdent(packageName));
		}
		LinkedListTree packageBlock = newBlock();
		pkg.addChildWithTokens(packageBlock);
		String className = typeNameFrom(qualifiedName);
		
		LinkedListTree clazz = synthesizeAS3Class(className);
		ASTUtils.addChildWithIndentation(packageBlock, clazz);
		return new AS3ASTCompilationUnit(unit);
	}

	public static ASCompilationUnit synthesizeInterface(String qualifiedName) {
		LinkedListTree unit = ASTUtils.newImaginaryAST(AS3Parser.COMPILATION_UNIT);
		LinkedListTree pkg = ASTUtils.newAST(AS3Parser.PACKAGE, "package");
		unit.addChildWithTokens(pkg);
		String packageName = packageNameFrom(qualifiedName);
		if (packageName != null) {
			pkg.addChildWithTokens(AS3FragmentParser.parseIdent(packageName));
		}
		LinkedListTree packageBlock = newBlock();
		pkg.addChildWithTokens(packageBlock);
		
		LinkedListTree iface = synthesizeAS3Interface(qualifiedName);
		packageBlock.addChildWithTokens(iface);
		return new AS3ASTCompilationUnit(unit);
	}

	private static LinkedListTree synthesizeAS3Interface(String qualifiedName) {
		LinkedListTree iface = ASTUtils.newImaginaryAST(AS3Parser.INTERFACE_DEF);
		LinkedListTree modifiers = ASTUtils.newImaginaryAST(AS3Parser.MODIFIERS);
		iface.addChildWithTokens(modifiers);
		modifiers.addChildWithTokens(ASTUtils.newAST(AS3Parser.PUBLIC, "public"));
		modifiers.appendToken(TokenBuilder.newSpace());
		iface.appendToken(TokenBuilder.newInterface());
		iface.appendToken(TokenBuilder.newSpace());
		iface.addChildWithTokens(ASTUtils.newAST(AS3Parser.IDENT, typeNameFrom(qualifiedName)));
		iface.appendToken(TokenBuilder.newSpace());
		LinkedListTree typeBlock = ASTUtils.newParentheticAST(AS3Parser.TYPE_BLOCK, AS3Parser.LCURLY, "{", AS3Parser.RCURLY, "}");
		iface.addChildWithTokens(typeBlock);
		LinkedListTree annos = ASTUtils.newPlaceholderAST(AS3Parser.ANNOTATIONS);
		iface.addChildWithTokens(0, annos);
		return iface;
	}

	private static String typeNameFrom(String qualifiedName) {
		int p = qualifiedName.lastIndexOf('.');
		if (p == -1) {
			return qualifiedName;
		}
		return qualifiedName.substring(p+1);
	}

	private static LinkedListTree synthesizeAS3Class(String className) {
		LinkedListTree clazz = ASTUtils.newImaginaryAST(AS3Parser.CLASS_DEF);
		LinkedListTree modifiers = ASTUtils.newImaginaryAST(AS3Parser.MODIFIERS);
		clazz.addChildWithTokens(modifiers);
		LinkedListTree modPublic = ASTUtils.newAST(AS3Parser.PUBLIC, "public");
		modifiers.addChildWithTokens(modPublic);
		modifiers.appendToken(TokenBuilder.newSpace());
		clazz.appendToken(TokenBuilder.newClass());
		clazz.appendToken(TokenBuilder.newSpace());
		clazz.addChildWithTokens(ASTUtils.newAST(AS3Parser.IDENT, className));
		clazz.appendToken(TokenBuilder.newSpace());
		LinkedListTree typeBlock = ASTUtils.newParentheticAST(AS3Parser.TYPE_BLOCK, AS3Parser.LCURLY, "{", AS3Parser.RCURLY, "}");
		LinkedListToken typeBlockNL = TokenBuilder.newNewline();
		typeBlock.getInitialInsertionAfter().afterInsert(typeBlockNL);
		typeBlock.setInitialInsertionAfter(typeBlockNL);
		clazz.addChildWithTokens(typeBlock);
		LinkedListTree annos = ASTUtils.newPlaceholderAST(AS3Parser.ANNOTATIONS);
		clazz.addChildWithTokens(0, annos);
		return clazz;
	}

	private static String packageNameFrom(String qualifiedName) {
		int p = qualifiedName.lastIndexOf('.');
		if (p == -1) {
			return null;
		}
		return qualifiedName.substring(0, p);
	}

	public static ASTASMethod newClassMethod(String name, Visibility visibility, String returnType) {
		LinkedListTree def = ASTUtils.newImaginaryAST(AS3Parser.METHOD_DEF);
		LinkedListTree decl = ASTUtils.newAST(AS3Parser.FUNCTION, "function");
		decl.appendToken(TokenBuilder.newSpace());
		def.addChildWithTokens(decl);
		LinkedListTree annos = ASTUtils.newPlaceholderAST(AS3Parser.ANNOTATIONS);
		def.addChildWithTokens(0, annos);
		def.addChildWithTokens(1, ModifierUtils.toModifiers(visibility));
		LinkedListTree methName = ASTUtils.newAST(AS3Parser.IDENT, name);
		def.addChildWithTokens(methName);
		def.addChildWithTokens(ASTUtils.newParentheticAST(AS3Parser.PARAMS, AS3Parser.LPAREN, "(", AS3Parser.RPAREN, ")"));
		if (returnType != null) {
			def.addChildWithTokens(AS3FragmentParser.parseTypeSpec(returnType));
		}
		def.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = newBlock();
		def.addChildWithTokens(block);

		return new ASTASMethod(def);
	}

	public static ASTASField newField(String name, Visibility visibility, String type) {
		if (name.indexOf('.') != -1) {
			throw new SyntaxException("field name must not contain '.'");
		}
		LinkedListTree decl = ASTUtils.newImaginaryAST(AS3Parser.VAR_DEF);
		LinkedListTree annos = ASTUtils.newPlaceholderAST(AS3Parser.ANNOTATIONS);
		decl.addChildWithTokens(annos);
		decl.addChildWithTokens(ModifierUtils.toModifiers(visibility));
		decl.addChildWithTokens(ASTUtils.newAST(AS3Parser.VAR, "var"));
		decl.appendToken(TokenBuilder.newSpace());
		LinkedListTree def = ASTUtils.newAST(AS3Parser.IDENT, name);
		decl.addChildWithTokens(def);
		if (type != null) {
			def.addChildWithTokens(AS3FragmentParser.parseTypeSpec(type));
		}
		decl.appendToken(TokenBuilder.newSemi());
		return new ASTASField(decl);
	}

	public static ASTASMethod newInterfaceMethod(String name, Visibility visibility, String returnType) {
		LinkedListTree def = ASTUtils.newImaginaryAST(AS3Parser.METHOD_DEF);
		LinkedListTree annos = ASTUtils.newPlaceholderAST(AS3Parser.ANNOTATIONS);
		def.addChildWithTokens(annos);
		def.addChildWithTokens(ModifierUtils.toModifiers(visibility));
		LinkedListTree decl = ASTUtils.newAST(AS3Parser.FUNCTION, "function");
		def.addChildWithTokens(decl);
		decl.appendToken(TokenBuilder.newSpace());
		LinkedListTree methName = ASTUtils.newAST(AS3Parser.IDENT, name);
		def.addChildWithTokens(methName);
		def.addChildWithTokens(ASTUtils.newParentheticAST(AS3Parser.PARAMS, AS3Parser.LPAREN, "(", AS3Parser.RPAREN, ")"));
		if (returnType != null) {
			def.addChildWithTokens(AS3FragmentParser.parseTypeSpec(returnType));
		}
		def.appendToken(TokenBuilder.newSemi());

		return new ASTASMethod(def);
	}

	public static LinkedListTree newBlock() {
		LinkedListTree ast = ASTUtils.newParentheticAST(AS3Parser.BLOCK,
		                                                AS3Parser.LCURLY, "{",
		                                                AS3Parser.RCURLY, "}");
		LinkedListToken nl = TokenBuilder.newNewline();
		ast.getInitialInsertionAfter().afterInsert(nl);
		ast.setInitialInsertionAfter(nl);
		return ast;
	}

	public static LinkedListTree newMetadataTag(String name) {
		LinkedListTree ast = ASTUtils.newParentheticAST(AS3Parser.ANNOTATION,
		                                                AS3Parser.LBRACK, "[",
		                                                AS3Parser.RBRACK, "]");
		ast.addChildWithTokens(ASTUtils.newAST(AS3Parser.IDENT, name));
		return ast;
	}
	
	public static LinkedListTree newIf(String condition) {
		LinkedListTree ifStmt = ASTUtils.newAST(AS3Parser.IF, "if");
		ifStmt.appendToken(TokenBuilder.newSpace());
		ifStmt.addChildWithTokens(AS3FragmentParser.parseCondition(condition));
		ifStmt.appendToken(TokenBuilder.newSpace());
		ifStmt.addChildWithTokens(ASTBuilder.newBlock());
		return ifStmt;
	}

	public static LinkedListTree newFor(String init, String condition, String iterate) {
		LinkedListTree forStmt = ASTUtils.newAST(AS3Parser.FOR, "for");
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.appendToken(TokenBuilder.newLParen());
		if (init != null) {
			forStmt.addChildWithTokens(AS3FragmentParser.parseForInit(init));
		} else {
			LinkedListTree initStmt = ASTUtils.newPlaceholderAST(AS3Parser.FOR_INIT);
			forStmt.addChildWithTokens(initStmt);
		}
		forStmt.appendToken(TokenBuilder.newSemi());
		forStmt.appendToken(TokenBuilder.newSpace());
		if (condition != null) {
			forStmt.addChildWithTokens(AS3FragmentParser.parseForCond(condition));
		} else {
			LinkedListTree condStmt = ASTUtils.newPlaceholderAST(AS3Parser.FOR_CONDITION);
			forStmt.addChildWithTokens(condStmt);
		}
		forStmt.appendToken(TokenBuilder.newSemi());
		forStmt.appendToken(TokenBuilder.newSpace());
		if (iterate != null) {
			forStmt.addChildWithTokens(AS3FragmentParser.parseForIter(iterate));
		} else {
			LinkedListTree iterStmt = ASTUtils.newPlaceholderAST(AS3Parser.FOR_ITERATOR);
			forStmt.addChildWithTokens(iterStmt);
		}
		forStmt.appendToken(TokenBuilder.newRParen());
		return forStmt;
	}

	public static LinkedListTree newForIn(String declaration, String expression) {
		LinkedListTree forStmt = ASTUtils.newAST(AS3Parser.FOR, "for");
		forStmt.appendToken(TokenBuilder.newSpace());
		genForInSetup(forStmt, declaration, expression);
		return forStmt;
	}

	public static LinkedListTree newForEachIn(String declaration, String expression) {
		LinkedListTree forStmt = ASTUtils.newAST(AS3Parser.FOR_EACH, "for");
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.appendToken(TokenBuilder.newEach());
		genForInSetup(forStmt, declaration, expression);
		return forStmt;
	}

	/**
	 * Common code for both for-in and for-each-in loop setup
	 */
	private static void genForInSetup(LinkedListTree forStmt, String declaration, String expression) {
		forStmt.appendToken(TokenBuilder.newLParen());
		forStmt.addChildWithTokens(AS3FragmentParser.parseForInVar(declaration));
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.appendToken(TokenBuilder.newIn());
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.addChildWithTokens(AS3FragmentParser.parseExpr(expression));
		forStmt.appendToken(TokenBuilder.newRParen());
	}

	public static LinkedListTree newWhile(String condition) {
		LinkedListTree whileStmt = ASTUtils.newAST(AS3Parser.WHILE, "while");
		whileStmt.appendToken(TokenBuilder.newSpace());
		whileStmt.addChildWithTokens(AS3FragmentParser.parseCondition(condition));
		return whileStmt;
	}

	public static LinkedListTree newDoWhile(String condition) {
		LinkedListTree doWhileStmt = ASTUtils.newAST(AS3Parser.DO, "do");
		doWhileStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTBuilder.newBlock();
		doWhileStmt.addChildWithTokens(block);
		doWhileStmt.appendToken(TokenBuilder.newSpace());
		doWhileStmt.appendToken(TokenBuilder.newWhile());
		doWhileStmt.appendToken(TokenBuilder.newSpace());
		doWhileStmt.addChildWithTokens(AS3FragmentParser.parseCondition(condition));
		doWhileStmt.appendToken(TokenBuilder.newSemi());
		return doWhileStmt;
	}

	public static LinkedListTree newSwitch(String condition) {
		LinkedListTree switchStmt = ASTUtils.newAST(AS3Parser.SWITCH, "switch");
		switchStmt.appendToken(TokenBuilder.newSpace());
		switchStmt.addChildWithTokens(AS3FragmentParser.parseCondition(condition));
		switchStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTBuilder.newBlock();
		switchStmt.addChildWithTokens(block);
		return switchStmt;
	}

	public static LinkedListTree newWith(String expr) {
		LinkedListTree withStmt = ASTUtils.newAST(AS3Parser.WITH, "with");
		withStmt.appendToken(TokenBuilder.newSpace());
		withStmt.addChildWithTokens(AS3FragmentParser.parseCondition(expr));
		return withStmt;
	}

	public static LinkedListTree newDeclaration(String assignment) {
		LinkedListTree declStmt = ASTUtils.newAST(AS3Parser.VAR, "var");
		declStmt.appendToken(TokenBuilder.newSpace());
		declStmt.addChildWithTokens(AS3FragmentParser.parseVariableDeclarator(assignment));
		declStmt.appendToken(TokenBuilder.newSemi());
		return declStmt;
	}

	public static LinkedListTree newReturn(String expr) {
		LinkedListTree returnStmt = ASTUtils.newAST(AS3Parser.RETURN, "return");
		if (expr != null) {
			returnStmt.appendToken(TokenBuilder.newSpace());
			returnStmt.addChildWithTokens(AS3FragmentParser.parseExpr(expr));
		}
		returnStmt.appendToken(TokenBuilder.newSemi());
		return returnStmt;
	}
}