/**
 * ASTStatementList.java
 * 
 * Copyright (c) 2006-2007 David Holroyd
 */

package uk.co.badgersinfoil.metaas.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import org.asdt.core.internal.antlr.AS3Parser;
import uk.co.badgersinfoil.metaas.dom.ASBlock;
import uk.co.badgersinfoil.metaas.dom.ASDeclarationStatement;
import uk.co.badgersinfoil.metaas.dom.ASDoWhileStatement;
import uk.co.badgersinfoil.metaas.dom.ASExpressionStatement;
import uk.co.badgersinfoil.metaas.dom.ASForEachInStatement;
import uk.co.badgersinfoil.metaas.dom.ASForInStatement;
import uk.co.badgersinfoil.metaas.dom.ASForStatement;
import uk.co.badgersinfoil.metaas.dom.ASIfStatement;
import uk.co.badgersinfoil.metaas.dom.ASReturnStatement;
import uk.co.badgersinfoil.metaas.dom.ASSwitchStatement;
import uk.co.badgersinfoil.metaas.dom.ASWhileStatement;
import uk.co.badgersinfoil.metaas.dom.ASWithStatement;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListToken;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTree;

// TODO: make the ASBlock implementation independant of the StatementContainer implementation

/**
 * Implements the behavior of StatementContainer.  May be managing statements
 * for either be a block, or a switch-statement label.
 */
public class ASTStatementList implements ASBlock {

	private LinkedListTree ast;

	public ASTStatementList(LinkedListTree ast) {
		this.ast = ast;
	}

	public void addStmt(String statement) {
		LinkedListTree stmt = AS3FragmentParser.parseStatement(statement);
		stmt.getStopToken().setNext(null);
		addStatement(stmt);
	}

	public ASExpressionStatement newExprStmt(String expr) {
		LinkedListTree stmt = AS3FragmentParser.parseExprStatement(expr);
		addStatement(stmt);
		return new ASTASExpressionStatement(stmt);
	}

	public void addComment(String text) {
		LinkedListToken comment = TokenBuilder.newSLComment("//" + text);
		String indent = findIndentForComment();
		ast.appendToken(TokenBuilder.newNewline());
		ast.appendToken(TokenBuilder.newWhiteSpace(indent));
		ast.appendToken(comment);
	}

	private String findIndentForComment() {
		LinkedListTree last = ast.getLastChild();
		String indent;
		if (last == null) {
			indent = "\t" + ASTUtils.findIndent(ast);
		} else {
			indent = ASTUtils.findIndent(last);
		}
		return indent;
	}

	public ASIfStatement newIf(String condition) {
		LinkedListTree ifStmt = ASTBuilder.newIf(condition);
		addStatement(ifStmt);
		return new ASTASIfStatement(ifStmt);
	}

	public ASForStatement newFor(String init, String condition, String iterate) {
		LinkedListTree forStmt = ASTBuilder.newFor(init, condition, iterate);
		addStatement(forStmt);
		appendBlock(forStmt);
		return new ASTASForStatement(forStmt);
	}

	public ASForInStatement newForIn(String declaration, String expression) {
		LinkedListTree forStmt = ASTBuilder.newForIn(declaration, expression);
		appendBlock(forStmt);
		addStatement(forStmt);
		return new ASTASForInStatement(forStmt);
	}

	public ASForEachInStatement newForEachIn(String declaration, String expression) {
		LinkedListTree forStmt = ASTBuilder.newForEachIn(declaration, expression);
		appendBlock(forStmt);
		addStatement(forStmt);
		return new ASTASForEachInStatement(forStmt);
	}

	public ASWhileStatement newWhile(String condition) {
		LinkedListTree whileStmt = ASTBuilder.newWhile(condition);
		appendBlock(whileStmt);
		addStatement(whileStmt);
		return new ASTASWhileStatement(whileStmt);
	}

	public ASDoWhileStatement newDoWhile(String condition) {
		LinkedListTree doWhileStmt = ASTBuilder.newDoWhile(condition);
		addStatement(doWhileStmt);
		return new ASTASDoWhileStatement(doWhileStmt);
	}

	public ASSwitchStatement newSwitch(String condition) {
		LinkedListTree switchStmt = ASTBuilder.newSwitch(condition);
		addStatement(switchStmt);
		return new ASTASSwitchStatement(findBlock(switchStmt));
	}

	public ASWithStatement newWith(String expr) {
		LinkedListTree withStmt = ASTBuilder.newWith(expr);
		appendBlock(withStmt);
		addStatement(withStmt);
		return new ASTASWithStatement(withStmt);
	}

	public ASDeclarationStatement newDeclaration(String assignment) {
		LinkedListTree declStmt = ASTBuilder.newDeclaration(assignment);
		addStatement(declStmt);
		return new ASTASDeclarationStatement(declStmt);
	}

	public ASReturnStatement newReturn(String expr) {
		LinkedListTree returnStmt = ASTBuilder.newReturn(expr);
		addStatement(returnStmt);
		return new ASTASReturnStatement(returnStmt);
	}

	private void addStatement(LinkedListTree stmt) {
		ASTUtils.addChildWithIndentation(ast, stmt);
	}

	public boolean containsCode() {
		return ast.getFirstChild() != null;
	}

	public LinkedListTree getAST() {
		return ast;
	}

	/**
	 * Appends a block node as a child of the given node (preceeding it
	 * with a space token) and returns a reference to the new block.
	 */
	private static LinkedListTree appendBlock(LinkedListTree ast) {
		ast.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTBuilder.newBlock();
		ast.addChildWithTokens(block);
		return block;
	}

	/**
	 * Returns the first BLOCK child of the given node
	 */
	private static LinkedListTree findBlock(LinkedListTree ast) {
		return (LinkedListTree)ast.getFirstChildWithType(AS3Parser.BLOCK);
	}

	public List getStatementList() {
		List result = new ArrayList();
		ASTIterator i = new ASTIterator(ast);
		while (i.hasNext()) {
			result.add(StatementBuilder.build(i.next()));
		}
		return Collections.unmodifiableList(result);
	}
}